(*******************************************************************************
  Copyright 2014-2015 MIT Lincoln Laboratory
  Project:     SPAR
  Author:      Alley Stoughton
  Description: Part of Private Count Retrieval (PCR) Protocol Proof
*******************************************************************************)

(* Inj.ec *)

(* Injective Maps Together with Their Domains and Ranges *)

prover "Alt-Ergo". timeout 5.  (* set prover and timeout *)

require import Option.
require import FMap.
require import FMapAux.
require import FSet.

pred Inj (mp : ('a, 'b)map, doma : 'a set, ran : 'b set) =
  dom mp = doma /\ rng mp = ran /\
  (forall (x1 x2 : 'a),
   mem x1 doma => mem x2 doma =>
   oget mp.[x1] = oget mp.[x2] =>
   x1 = x2).

lemma InjIntro (mp : ('a, 'b)map, doma : 'a set, ran : 'b set) :
  dom mp = doma /\ rng mp = ran /\
  (forall (x1 : 'a, x2 : 'a),
   mem x1 doma => mem x2 doma =>
   oget mp.[x1] = oget mp.[x2] =>
   x1 = x2) =>
  Inj mp doma ran
by trivial.

lemma InjElim (mp : ('a, 'b)map, doma : 'a set, ran : 'b set) :
  Inj mp doma ran =>
  dom mp = doma /\ rng mp = ran /\
  (forall (x1 : 'a, x2 : 'a),
   mem x1 doma => mem x2 doma =>
   oget mp.[x1] = oget mp.[x2] =>
   x1 = x2).
proof.
intros H1; trivial.
qed.

lemma InjEmpty :
  Inj FMap.empty<:'a,'b> FSet.empty<:'a> FSet.empty<:'b>
by smt.

lemma InjAdd (mp : ('a, 'b)map, doma : 'a set, ran : 'b set, x : 'a, y : 'b) :
  Inj mp doma ran => ! mem x doma => ! mem y ran =>
  Inj mp.[x <- y] (add x doma) (add y ran).
proof.
intros H1 H2 H3.
elim H1 => H4 [H5 H6].
split; [smt | split;smt].
qed.
