(*******************************************************************************
  Copyright 2014-2015 MIT Lincoln Laboratory
  Project:     SPAR
  Author:      Alley Stoughton
  Description: Part of Private Count Retrieval (PCR) Protocol Proof
*******************************************************************************)

(* ListAux.ec *)

(* Auxiliary Operators and Lemmas for Lists *)

prover "Alt-Ergo". timeout 5.  (* set prover and timeout *)

require import Option.
require import Int.
require import List.

lemma RevLen (xs : 'a list) :
  `|rev xs| = `|xs|.
proof.
elim xs; smt.
qed.

lemma NthRev (xs : 'a list) (i : int) :
  0 <= i < `|xs| =>
  oget(nth (rev xs) (`|xs| - (i + 1))) = oget(nth xs i).
proof.
move : i.
elim xs.
(* basis step *)
smt.
(* inductive step *)
intros x xs IH i H1.
cut -> : rev(x :: xs) = rev xs ++ [x] by smt.
case (i = 0) => H2.
smt.
cut H3 : 0 <= i - 1 < `|xs| by smt.
cut -> : `|x :: xs| - (i + 1) = `|xs| - i by smt.
cut H4 : 0 <= `|xs| - i < `|xs| by smt.
cut -> :
  oget(nth (rev xs ++ [x]) (`|xs| - i)) =
  oget(nth (rev xs) (`|xs| - i)) by smt.
cut -> :
  oget(if i = 0 then Some x else nth xs (i - 1)) =
  oget(nth xs (i - 1)) by smt.
cut {1} -> : i = (i - 1) + 1 by smt.
apply IH => //.
qed.

lemma MapNth (xs : 'a list, f : 'a -> 'b, i : int) :
  0 <= i < `|xs| =>
  oget(nth (map f xs) i) = f(oget(nth xs i)).
proof.
move : i.
elim xs => i H1.
(* basis step *)
simplify; smt.
(* inductive step *)
intros IH i0 [H2 H3].
simplify.
case (i0 = 0) => H4.
smt.
(* i0 >= 1 *)
cut H5 : i0 >= 1 by smt.
rewrite IH.
split.
smt.
intros H6.
cut H7 : `|i :: H1| = `|H1| + 1 by smt.
smt.
reflexivity.
qed.

(* if 0 <= i < `|xs|, then del xs i deletes the ith element of xs,
   returning the remaining elements in the same order *)

op del (xs : 'a list) (i : int) : 'a list =
  with xs = "[]"      => []  (* error case *)
  with xs = (::) x xs =>
    if i = 0
    then xs
    else x :: del xs (i - 1).

lemma DelLen (xs : 'a list) (i : int) :
  0 <= i < `|xs| =>
  `|del xs i| = `|xs| - 1.
proof.
move : i.
elim xs; smt.
qed.

lemma DelConsLeft (x : 'a) (xs : 'a list) (i : int) :
  del (x :: xs) 0 = xs
by smt.

lemma DelConsRight (x : 'a) (xs : 'a list) (i : int) :
  0 <= i < `|xs| =>
  del (x :: xs) (1 + i) = x :: del xs i
by smt.

lemma DelConcatLeft (xs ys : 'a list) (i : int) :
  0 <= i < `|xs| =>
  del (xs ++ ys) i = del xs i ++ ys.
proof.
move : i.
elim xs; smt.
qed.

lemma DelConcatRight (xs ys : 'a list) (i : int) :
  0 <= i < `|ys| =>
  del (xs ++ ys) (`|xs| + i) = xs ++ del ys i.
proof.
elim xs; smt.
qed.
