(*******************************************************************************
  Copyright 2014-2015 MIT Lincoln Laboratory
  Project:     SPAR
  Author:      Alley Stoughton
  Description: Part of Private Count Retrieval (PCR) Protocol Proof
*******************************************************************************)

(* NumOccs.ec *)

(* Operators for Counting Numbers of Occurrences of Elements in Lists *)

prover "Alt-Ergo". timeout 10.  (* set prover and timeout *)

require import Option.
require import Int.
require import List.
require import ListAux.

(* if 0 <= n <= `|ys|, then numOccsUpTo x ys n is the number of
   occurrences of x in the first n elements of ys *)

op numOccsUpTo (x : 'a) (ys : 'a list) (n : int) : int =
  with ys = "[]"      => 0
  with ys = (::) y ys =>
    if n = 0
    then 0
    else (if x = y then 1 else 0) + numOccsUpTo x ys (n - 1).

lemma NumOccsUpTo_Zero (x : 'a, ys : 'a list) :
  numOccsUpTo x ys 0 = 0.
proof.
elim ys; smt.
qed.

lemma NumOccsUpTo_NonZeroEqual (x : 'a, ys : 'a list, i : int) :
  0 <= i < `|ys| =>
  oget(nth ys i) = x =>
  numOccsUpTo x ys (i + 1) = numOccsUpTo x ys i + 1.
proof.
move : i.
elim ys.
(* basis step *)
smt.
(* inductive step *)
intros y ys IH i H1 H2.
case (i = 0) => H3.
cut H4 : y = x; smt.
(* i <> 0 *)
cut H4 : 0 <= i - 1 < `|ys| by smt.
cut -> :
  (if i + 1 = 0 then 0 else (if x = y then 1 else 0) +
   numOccsUpTo x ys (i + 1 - 1)) =
  ((if x = y then 1 else 0) + numOccsUpTo x ys (i + 1 - 1)) by smt.
cut -> :
  (if i = 0 then 0 else (if x = y then 1 else 0) + numOccsUpTo x ys (i - 1)) =
  ((if x = y then 1 else 0) + numOccsUpTo x ys (i - 1)) by smt.
cut -> : (i + 1 - 1) = i by smt.
cut H5 : oget(nth ys (i - 1)) = x; smt.
qed.

lemma NumOccsUpTo_NonZeroNotEqual (x : 'a, ys : 'a list, i : int) :
  0 <= i < `|ys| =>
  oget(nth ys i) <> x =>
  numOccsUpTo x ys (i + 1) = numOccsUpTo x ys i.
proof.
move : i.
elim ys.
(* basis step *)
smt.
(* inductive step *)
intros y ys IH i H1 H2.
case (i = 0) => H3.
cut H4 : y <> x; smt.
(* i <> 0 *)
cut H4 : 0 <= i - 1 < `|ys| by smt.
cut -> :
  (if i + 1 = 0 then 0
   else (if x = y then 1 else 0) + numOccsUpTo x ys (i + 1 - 1)) =
  ((if x = y then 1 else 0) + numOccsUpTo x ys (i + 1 - 1)) by smt.
cut -> :
  (if i = 0 then 0 else (if x = y then 1 else 0) + numOccsUpTo x ys (i - 1)) =
  ((if x = y then 1 else 0) + numOccsUpTo x ys (i - 1)) by smt.
cut -> : (i + 1 - 1) = i by smt.
cut H5 : oget(nth ys (i - 1)) <> x; smt.
qed.

lemma NumOccsUpTo_RangeNonNeg (x : 'a, ys : 'a list, i : int) :
  0 <= i <= `|ys| => 0 <= numOccsUpTo x ys i.
proof.
cut H : forall (j : int), 0 <= j => j <= `|ys| => 0 <= numOccsUpTo x ys j.
  intros j; elim j; smt.
smt.
qed.

lemma NumOccsUpTo_ZeroOverallImpliesZeroAllButLast
  (x : 'a, ys : 'a list, i : int) :
  0 <= i < `|ys| =>
  numOccsUpTo x ys (i + 1) = 0 =>
  numOccsUpTo x ys i = 0
by smt.

lemma NumOccsUpTo_ZeroOverallImpliesNotEqualLast
  (x : 'a, ys : 'a list, i : int) :
  0 <= i < `|ys| =>
  numOccsUpTo x ys (i + 1) = 0 =>
  oget(nth ys i) <> x.
proof.
intros [H1 H2] H3.
cut H4 : numOccsUpTo x ys i = 0
  by apply NumOccsUpTo_ZeroOverallImpliesZeroAllButLast => //.
case (oget (nth ys i) = x) => H5.
(* oget (nth ys i) = x -- need a contradiction *)
cut H6 : numOccsUpTo x ys (i + 1) = numOccsUpTo x ys i + 1
  by apply NumOccsUpTo_NonZeroEqual.
smt.
(* oget (nth ys i) <> x *)
trivial.
qed.

lemma NumOccsUpTo_DropLastPart (x : 'a, ys, zs : 'a list) :
  numOccsUpTo x (ys ++ zs) `|ys| = numOccsUpTo x ys `|ys|.
proof.
cut H :
  forall (i : int),
  0 <= i => i <= `|ys| =>
  numOccsUpTo x (ys ++ zs) i = numOccsUpTo x ys i.
  intros i; elim i.
  (* basis step *)
  smt.
  (* inductive step *)
  intros i H1 IH H2.
  cut H3 : i <= `|ys| by smt.
  cut H4 : numOccsUpTo x (ys ++ zs) i = numOccsUpTo x ys i
    by apply IH.
  case (oget (nth ys i) = x) => H5.
  (* oget (nth ys i) = x *)
  cut H6 : oget(nth (ys ++ zs) i) = x by smt.
  cut -> :
    numOccsUpTo x (ys ++ zs) (i + 1) =
    numOccsUpTo x (ys ++ zs) i + 1.
    apply NumOccsUpTo_NonZeroEqual; [smt | assumption].
  cut -> :
    numOccsUpTo x ys (i + 1) =
    numOccsUpTo x ys i + 1.
    apply NumOccsUpTo_NonZeroEqual; [smt | assumption].
  rewrite H4; reflexivity.
  (* oget (nth ys i) <> x *)
  cut -> :
    numOccsUpTo x (ys ++ zs) (i + 1) =
    numOccsUpTo x (ys ++ zs) i.
    apply NumOccsUpTo_NonZeroNotEqual; smt.
  cut -> :
    numOccsUpTo x ys (i + 1) =
    numOccsUpTo x ys i.
    apply NumOccsUpTo_NonZeroNotEqual; smt.
  rewrite H4; reflexivity.
apply H; [smt | trivial].
qed.

(* numOccs x ys is the number of occurrences of x in ys *)

op numOccs (x : 'a) (ys : 'a list) : int = numOccsUpTo x ys `|ys|.

lemma NumOccs_NotMemImpliesZero (x : 'a, ys : 'a list) :
  ! mem x ys => numOccs x ys = 0.
proof.
intros Mem.
cut H :
  forall (i : int),
  0 <= i => i <= `|ys| =>
  numOccsUpTo x ys i = 0.
  intros i; elim i.
  (* basis step *)
  intros H1; apply NumOccsUpTo_Zero.
  (* inductive step *)
  intros i H1 IH H2.
  case (oget(nth ys i) = x) => H3.
  (* oget(nth ys i) = x *)
  rewrite NumOccsUpTo_NonZeroEqual.
  smt. assumption.
  rewrite IH; smt.
  (* oget(nth ys i) <> x *)
  rewrite NumOccsUpTo_NonZeroNotEqual.
  smt. assumption.
  rewrite IH; [smt | reflexivity].
rewrite /numOccs.
apply (H `|ys|); [smt | trivial].
qed.

lemma NumOccs_ZeroImpliesNotMem (x : 'a, ys : 'a list) :
  numOccs x ys = 0 => ! mem x ys.
proof.
cut H1 :
  forall (i : int),
  0 <= i => i <= `|ys| =>
  numOccsUpTo x ys i = 0 =>
  !(exists (j : int),
    0 <= j /\ j < i /\ oget(nth ys j) = x).
  intros i; elim i.
  (* basis step *)
  smt.
  (* inductive step *)
  intros i IH H1 H2 H3.
  cut H4 : numOccsUpTo x ys i = 0.
    apply NumOccsUpTo_ZeroOverallImpliesZeroAllButLast;
    [smt | assumption].
  cut H5 : oget(nth ys i) <> x.
    apply NumOccsUpTo_ZeroOverallImpliesNotEqualLast;
    [smt | assumption].
  cut H6 :
    !(exists (j : int), 0 <= j /\ j < i /\ oget (nth ys j) = x).
    apply H1; [smt | assumption].
  smt.
intros H2.
cut H3 : numOccsUpTo x ys `|ys| = 0; first trivial.
cut H4 : !(exists (j : int), 0 <= j /\ j < `|ys| /\ oget (nth ys j) = x).
  apply H1; [smt | trivial | assumption].
cut H5 : mem x ys \/ ! mem x ys by smt.
elim H5.
intros H5.
cut H6 : exists (j : int), 0 <= j /\ j < `|ys| /\ nth ys j = Some x by smt.
(* have contradiction *)
smt.
(* trivial case *)
trivial.
qed.

lemma NumOccsCons (x y : 'a) (zs : 'a list) :
  numOccs x (y :: zs) =
  (if x = y then 1 else 0) + numOccs x zs
by smt.

lemma NumOccsConcat (x : 'a) (ys zs : 'a list) :
  numOccs x (ys ++ zs) = numOccs x ys + numOccs x zs.
proof.
elim ys; smt.
qed.

lemma NumOccsAddEnd (x y : 'a) (zs : 'a list) :
  numOccs x (zs ++ [y]) =
  numOccs x zs + (if x = y then 1 else 0).
proof.
rewrite (NumOccsConcat x zs [y]).
rewrite (NumOccsCons x y []).
reflexivity.
qed.

lemma NumOccsRev (x : 'a) (ys : 'a list) :
  numOccs x (rev ys) = numOccs x ys.
proof.
elim ys; smt.
qed.

lemma NumOccsDel (i : int) (z : 'a) (xs : 'a list) :
  0 <= i < `|xs| =>
  numOccs z (del xs i) + (if z = oget(nth xs i) then 1 else 0) =
  numOccs z xs.
proof.
move : i.
elim xs; smt.
qed.

lemma NumOccsDelAdd (i : int) (z : 'a) (xs ys : 'a list) :
  0 <= i < `|xs| =>
  numOccs z (del xs i) + numOccs z (ys ++ [oget(nth xs i)]) =
  numOccs z xs + numOccs z ys
by smt.

lemma NumOccsUpTo_Rev_Zero (x : 'a, ys : 'a list) :
  numOccsUpTo x (rev ys) 0 = 0
by smt.

lemma NumOccsUpTo_Rev_NonZeroEqual (x : 'a, ys : 'a list, i : int) :
  0 <= i < `|ys| =>
  oget(nth ys i) = x =>
  numOccsUpTo x (rev ys) (`|ys| - i) =
  numOccsUpTo x (rev ys) (`|ys| - (i + 1)) + 1
by smt.

lemma NumOccsUpTo_Rev_NonZeroNotEqual (x : 'a, ys : 'a list, i : int) :
  0 <= i < `|ys| =>
  oget(nth ys i) <> x =>
  numOccsUpTo x (rev ys) (`|ys| - i) =
  numOccsUpTo x (rev ys) (`|ys| - (i + 1))
by smt.
