(*******************************************************************************
  Copyright 2014-2015 MIT Lincoln Laboratory
  Project:     SPAR
  Author:      Alley Stoughton
  Description: Part of Private Count Retrieval (PCR) Protocol Proof
*******************************************************************************)

(* RealAux.ec *)

(* Auxiliary Lemmas for Reals *)

prover "Alt-Ergo". timeout 5.  (* set prover and timeout *)

require import Real.

lemma RealMultLeftLEq (x y z : real) :
  x <= y => z >= 0%r => x * z <= y * z.
proof.
intros H1 H2.
apply mulrMle; smt.
qed.

lemma RealMultRightLEq (x y z : real) :
  x <= y => z >= 0%r => z * x <= z * y.
proof.
intros H1 H2.
cut -> : z * x = x * z by smt.
cut -> : z * y = y * z by smt.
apply RealMultLeftLEq => //.
qed.

lemma RealDivToMultTimesInv (x y : real) :
  y <> 0%r => x / y = x * (1%r / y)
by trivial.

lemma RealInvGTZero (x : real) :
  x > 0%r => 1%r / x > 0%r.
proof.
intros H1.
cut -> : 1%r / x = inv x by smt.
cut H2 : 0%r < inv x.
  apply sign_inv; smt.
smt.
qed.

lemma RealDivTopLEq (x y z : real) :
  x <= y => z > 0%r => x / z <= y / z
by smt.

lemma RealInvLEq (x y : real) :
  0%r < x <= y => 1%r / y <= 1%r / x.
proof.
intros [H1 H2].
cut -> : 1%r / y = x / (x * y).
  cut {1} -> : x = x * 1%r by smt.
  cut -> : x * 1%r / (x * y) = (x / x) * (1%r / y) by smt.
  cut -> : x / x = 1%r by smt.
  smt.
cut -> : 1%r / x = y / (y * x).
  cut {1} -> : y = y * 1%r by smt.
  cut -> : y * 1%r / (y * x) = (y / y) * (1%r / x) by smt.
  cut -> : y / y = 1%r by smt.
  smt.
cut -> : x * y = y * x by smt.
apply RealDivTopLEq => //.
cut H3 : 0%r * y < x * y; first apply mulrM; smt.
smt.
qed.

lemma RealDivBotLEq (x y z : real) :
  0%r < x <= y => z >= 0%r => z / y <= z / x.
proof.
intros [H1 H2] H3.
cut -> : z / y = z * (1%r / y).
  apply RealDivToMultTimesInv.
cut -> : z / x = z * (1%r / x).
  apply RealDivToMultTimesInv.
apply RealMultRightLEq;
  [apply RealInvLEq; trivial |
   assumption].
qed.

lemma RealFracSumEqualDenom (x y z : real) :
  z <> 0%r =>
  x / z + y / z = (x + y) / z
by smt.

lemma RealDistrRight (x y z : real) :
  (x + y) * z = x * z + y * z
by smt.

lemma RealDistrLeft (x y z : real) :
  z * (x + y) = z * x + z * y
by smt.
