(*******************************************************************************
  Copyright 2014-2015 MIT Lincoln Laboratory
  Project:     SPAR
  Author:      Alley Stoughton
  Description: Part of Private Count Retrieval (PCR) Protocol Proof
*******************************************************************************)

(* SumAux.ec *)

(* Auxiliary Lemmas for Summations *)

prover "Alt-Ergo". timeout 5.  (* set prover and timeout *)

require import Int.
require import Real.
require import RealAux.
require import Sum.

lemma SummationFunFromZeroToMinusOne (f : int -> real) :
  int_sum f (intval 0 (0 - 1)) = 0%r
by smt.

lemma SummationFunFromZeroToNonNeg (f : int -> real, n : int) :
  0 <= n =>
  int_sum f (intval 0 n) = int_sum f (intval 0 (n - 1)) + f n.
proof.
intros H1; rewrite /int_sum /intval; smt.
qed.

lemma SummationFunFromZeroToZero (f : int -> real) :
  int_sum f (intval 0 0) = f 0.
proof.
rewrite (SummationFunFromZeroToNonNeg f 0) => //.
rewrite (SummationFunFromZeroToMinusOne f).
trivial.
qed.

lemma SummationFromZeroToNMinusOne (n : int) :
  0 <= n =>
  int_sum
  (fun i, i%r)
  (intval 0 (n - 1)) = (n - 1)%r * n%r / 2%r.
proof.
elim n.
(* basis step *)
apply (SummationFunFromZeroToMinusOne (fun i, i%r)).
(* inductive step *)
intros i H1 IH.
cut -> : i + 1 - 1 = i by smt.
case (i = 0) => H2.
(* i = 0 *)
rewrite H2.
rewrite (SummationFunFromZeroToZero (fun i, i%r)).
trivial.
(* i >= 1 *)
cut H3 : i >= 1 by smt.
cut -> :
  int_sum (fun i, i%r) (intval 0 i) =
  int_sum (fun i, i%r) (intval 0 (i - 1)) + i%r.
  apply (SummationFunFromZeroToNonNeg (fun i, i%r) i); trivial.
rewrite IH.
(* goal :
H1: 0 <= i
IH: int_sum (lambda (i0 : int), i0%r) (intval 0 (i - 1)) =
    (i - 1)%r * i%r / 2%r
H2: ! i = 0
H3: i >= 1
------------------------------------------------------------------------
(i - 1)%r * i%r / 2%r + i%r = i%r * (i + 1)%r / 2%r
*)
clear H1 IH H2 H3.
cut {2} -> : i%r = (2%r * i%r) / (2%r * 1%r).
  cut -> : (2%r * i%r) / (2%r * 1%r) = (2%r / 2%r) * (i%r / 1%r) by smt.
  cut -> : 2%r / 2%r = 1%r by smt.
  cut -> : i%r / 1%r = i%r by smt.
  smt.
cut -> : 2%r * 1%r = 2%r by smt.
cut -> :
  (i - 1)%r * i%r / 2%r + (2%r * i%r) / 2%r =
  (((i - 1)%r * i%r) + (2%r * i%r)) / 2%r
  by apply RealFracSumEqualDenom.
cut -> : (i - 1)%r * i%r + 2%r * i%r = i%r * (i + 1)%r.
  cut -> : (i - 1)%r * i%r = i%r * i%r + -i%r by smt.
  cut H2 : -i%r + 2%r * i%r = i%r by smt.
  cut -> :
    i%r * i%r + -i%r + 2%r * i%r =
    i%r * i%r + i%r by smt.
  smt.
reflexivity.
qed.

lemma SummationFromZeroToNMinusOneMultRight (x : real, n : int) :
  0 <= n =>
  int_sum (fun i, i%r * x) (intval 0 (n - 1)) =
  int_sum (fun i, i%r) (intval 0 (n - 1)) * x.
proof.
elim n.
(* basis step *)
cut -> : int_sum (fun (i : int), i%r * x) (intval 0 (0 - 1)) = 0%r
  by apply SummationFunFromZeroToMinusOne.
cut -> : int_sum (fun (i : int), i%r) (intval 0 (0 - 1)) = 0%r
  by apply SummationFunFromZeroToMinusOne.
trivial.
(* inductive step *)
intros i H1 IH.
cut -> : i + 1 - 1 = i by smt.
case (i = 0) => H2.
(* i = 0 *)
rewrite H2.
rewrite (SummationFunFromZeroToZero (fun i, i%r * x)).
rewrite (SummationFunFromZeroToZero (fun i, i%r)).
trivial.
(* i >= 1 *)
cut H3 : i >= 1 by smt.
cut -> :
  int_sum (fun i, i%r * x) (intval 0 i) =
  int_sum (fun i, i%r * x) (intval 0 (i - 1)) + i%r * x.
  apply (SummationFunFromZeroToNonNeg (fun i, i%r * x) i); trivial.
cut -> :
  int_sum (fun i, i%r) (intval 0 i) =
  int_sum (fun i, i%r) (intval 0 (i - 1)) + i%r.
  apply (SummationFunFromZeroToNonNeg (fun i, i%r) i); trivial.
rewrite IH.
rewrite SummationFromZeroToNMinusOne; trivial.
(* goal:
(i - 1)%r * i%r / 2%r * x + i%r * x = ((i - 1)%r * i%r / 2%r + i%r) * x
*)
rewrite - RealDistrRight => //.
qed.
